@@ -9,6 +9,9 @@ JavaVM *jvm;
 /* this will be set when Java tries to exit() but we carry on */
 int java_is_dead = 0;
 
+/* current JVM state */
+int rJava_JVM_state = JVM_STATE_NONE;
+
 /* cached, global objects */
 
 jclass javaStringClass;
@@ -40,10 +43,10 @@ int rJava_initialized = 0;
 static int    jvm_opts=0;
 static char **jvm_optv=0;
 
-#ifdef JNI_VERSION_1_2 
+#ifdef JNI_VERSION_1_2
 static JavaVMOption *vm_options;
 static JavaVMInitArgs vm_args;
-#else  
+#else
 #error "Java/JNI 1.2 or higher is required!"
 #endif
 
@@ -78,6 +81,7 @@ static int JNICALL vfprintf_hook(FILE *f, const char *fmt, va_list ap) {
 static void JNICALL exit_hook(int status) {
     /* REprintf("\nJava requested System.exit(%d), trying to raise R error - this may crash if Java is in a bad state.\n", status); */
     java_is_dead = 1;
+    rJava_JVM_state = JVM_STATE_DEAD;
     Rf_error("Java called System.exit(%d) requesting R to quit - trying to recover", status);
     /* FIXME: we could do something smart here such as running a call-back
        into R ... jump into R event loop ... at any rate we cannot return,
@@ -85,6 +89,12 @@ static void JNICALL exit_hook(int status) {
     exit(status);
 }
 
+int existingJVMs(void) {
+    jsize vms = 0;
+    JavaVM *jvms[32];
+    return (JNI_GetCreatedJavaVMs(jvms, 32, &vms) >= 0) ? vms : 0;
+}
+
 /* in reality WIN64 implies WIN32 but to make sure ... */
 #if defined(_WIN64) || defined(_WIN32)
 #include <io.h>
@@ -98,19 +108,19 @@ static void JNICALL exit_hook(int status) {
      JVM version 10 and above on Linux; only used with JVM_STACK_WORKAROUND */
 static int initJVM(const char *user_classpath, int opts, char **optv, int hooks,
                    int disableGuardPages) {
-  int total_num_properties, propNum = 0;
+  int total_num_properties, propNum = 0, add_Xss = 1;
   jint res;
   char *classpath;
-  
+
   if(!user_classpath)
     /* use the CLASSPATH environment variable as default */
     user_classpath = getenv("CLASSPATH");
   if(!user_classpath) user_classpath = "";
-  
+
   vm_args.version = JNI_VERSION_1_2;
   if(JNI_GetDefaultJavaVMInitArgs(&vm_args) != JNI_OK) {
     error("JNI 1.2 or higher is required");
-    return -1;    
+    return -1;
   }
 
   vm_args.version = JNI_VERSION_1_2; /* should we do that or keep the default? */
@@ -132,23 +142,31 @@ static int initJVM(const char *user_classpath, int opts, char **optv, int hooks,
 
   /* leave room for class.path, and optional jni args */
   total_num_properties = 8 + opts;
-    
+
   vm_options = (JavaVMOption *) calloc(total_num_properties, sizeof(JavaVMOption));
   vm_args.options = vm_options;
   vm_args.ignoreUnrecognized = disableGuardPages ? JNI_FALSE : JNI_TRUE;
-  
+
   classpath = (char*) calloc(24 + strlen(user_classpath), sizeof(char));
-  sprintf(classpath, "-Djava.class.path=%s", user_classpath);
-  
-  vm_options[propNum++].optionString = classpath;   
-  
+  if (!classpath)
+      error("Cannot allocate memory for classpath");
+  snprintf(classpath, (24 + strlen(user_classpath)) * sizeof(char), "-Djava.class.path=%s", user_classpath);
+
+  vm_options[propNum++].optionString = classpath;
+
   /*   print JNI-related messages */
   /* vm_options[propNum++].optionString = "-verbose:class,jni"; */
-  
+
   if (optv) {
-    int i=0;
-    while (i<opts) {
-      if (optv[i]) vm_options[propNum++].optionString = optv[i];
+    int i = 0;
+    while (i < opts) {
+      if (optv[i]) {
+	vm_options[propNum++].optionString = optv[i];
+	/* we look for -Xss since we want to raise it to avoid overflows,
+	   but not if the user already supplied it */
+	if (!strncmp(optv[i], "-Xss", 4))
+	  add_Xss = 0;
+      }
       i++;
     }
   }
@@ -164,19 +182,23 @@ static int initJVM(const char *user_classpath, int opts, char **optv, int hooks,
     vm_options[propNum++].optionString = "-XX:+UnlockExperimentalVMOptions";
     vm_options[propNum++].optionString = "-XX:+DisablePrimordialThreadGuardPages";
   }
+  if (add_Xss)
+    vm_options[propNum++].optionString = "-Xss2m";
+
   vm_args.nOptions = propNum;
 
   /* Create the Java VM */
   res = JNI_CreateJavaVM(&jvm,(void **)&eenv, &vm_args);
-  
+
   if (disableGuardPages && (res != 0 || !eenv))
     return -2; /* perhaps this VM does not allow disabling guard pages */
 
   if (res != 0)
     error("Cannot create Java virtual machine (%d)", res);
   if (!eenv)
-    error("Cannot obtain JVM environemnt");
+    error("Cannot obtain JVM environment");
 
+  rJava_JVM_state = JVM_STATE_CREATED;
 #if defined(_WIN64) || defined(_WIN32)
   _setmode(0, _O_TEXT);
 #endif
@@ -208,6 +230,8 @@ static void *initJVMthread(void *classpath)
   thInitResult=initJVM((char*)classpath, jvm_opts, jvm_optv, default_hooks, 0);
   if (thInitResult) return 0;
 
+  rJava_JVM_state = JVM_STATE_CREATED;
+
   init_rJava();
 
   lenv = eenv; /* we make a local copy before unlocking just in case
@@ -237,8 +261,8 @@ HIDE void init_rJava(void) {
   jclass c;
   JNIEnv *env = getJNIEnv();
   if (!env) return; /* initJVMfailed, so we cannot proceed */
-  
-  /* get global classes. we make the references explicitely global (although unloading of String/Object is more than unlikely) */
+
+  /* get global classes. we make the references explicitly global (although unloading of String/Object is more than unlikely) */
   c=(*env)->FindClass(env, "java/lang/String");
   if (!c) error("unable to find the basic String class");
   javaStringClass=(*env)->NewGlobalRef(env, c);
@@ -265,17 +289,17 @@ HIDE void init_rJava(void) {
 
   mid_forName  = (*env)->GetStaticMethodID(env, javaClassClass, "forName", "(Ljava/lang/String;ZLjava/lang/ClassLoader;)Ljava/lang/Class;");
   if (!mid_forName) error("cannot obtain Class.forName method ID");
-  
+
   mid_getName  = (*env)->GetMethodID(env, javaClassClass, "getName", "()Ljava/lang/String;");
   if (!mid_getName) error("cannot obtain Class.getName method ID");
-  
+
   mid_getSuperclass =(*env)->GetMethodID(env, javaClassClass, "getSuperclass", "()Ljava/lang/Class;");
   if (!mid_getSuperclass) error("cannot obtain Class.getSuperclass method ID");
-  
+
   mid_getField = (*env)->GetMethodID(env, javaClassClass, "getField",
 				     "(Ljava/lang/String;)Ljava/lang/reflect/Field;");
   if (!mid_getField) error("cannot obtain Class.getField method ID");
- 
+
   mid_getType  = (*env)->GetMethodID(env, javaFieldClass, "getType",
 				     "()Ljava/lang/Class;");
   if (!mid_getType) error("cannot obtain Field.getType method ID");
@@ -294,10 +318,10 @@ static SEXP RinitJVM_real(SEXP par, int disableGuardPages)
   int r=0;
   JavaVM *jvms[32];
   jsize vms=0;
-  
+
   jvm_opts=0;
   jvm_optv=0;
-  
+
   if (TYPEOF(e)==STRSXP && LENGTH(e)>0)
 	  c=CHAR(STRING_ELT(e,0));
 
@@ -337,7 +361,7 @@ static SEXP RinitJVM_real(SEXP par, int disableGuardPages)
   }
   if (jvm_opts)
       jvm_optv[jvm_opts] = 0;
-  
+
   r=JNI_GetCreatedJavaVMs(jvms, 32, &vms);
   if (r) {
     Rf_error("JNI_GetCreatedJavaVMs returned %d\n", r);
@@ -348,7 +372,10 @@ static SEXP RinitJVM_real(SEXP par, int disableGuardPages)
       while (i<vms) {
 	if (jvms[i]) {
 	  if (!(*jvms[i])->AttachCurrentThread(jvms[i], (void**)&eenv, NULL)) {
-            _dbg(rjprintf("RinitJVM: Attached to existing JVM #%d.\n", i+1));
+	      /* attaching our own created JVM doesn't change it to attached */
+	      if (rJava_JVM_state != JVM_STATE_CREATED)
+		  rJava_JVM_state = JVM_STATE_ATTACHED;
+	      _dbg(rjprintf("RinitJVM: Attached to existing JVM #%d.\n", i+1));
 	    break;
 	  }
 	}
@@ -380,6 +407,8 @@ static SEXP RinitJVM_real(SEXP par, int disableGuardPages)
   _dbg(rjprintf("RinitJVM(threads): attach\n"));
   /* since JVM was initialized by another thread, we need to attach ourselves */
   (*jvm)->AttachCurrentThread(jvm, (void**)&eenv, NULL);
+  if (rJava_JVM_state != JVM_STATE_CREATED)
+      rJava_JVM_state = JVM_STATE_ATTACHED;
   _dbg(rjprintf("RinitJVM(threads): done.\n"));
   r = thInitResult;
 #else
@@ -468,7 +497,7 @@ extern int R_CStackDir;
   bound is the new first address not to access
   dir is 1 (stack grows up) or -1 (stack grows down, the usual)
     so, incrementing by dir one traverses from stack start, from the oldest
-    things on the stace; note that in R_CStackDir, 1 means stack grows down
+    things on the stack; note that in R_CStackDir, 1 means stack grows down
 
   returns NULL in case of error, limit when no new limit is found,
     a new limit when found
@@ -503,7 +532,7 @@ static char* findBound(char *from, char *limit, int dir)
     _exit(0);
   } else {
     /* Parent process, writes data to the pipe looking for EFAULT error which
-       indicates an inaccesible page. For performance, the reading is first
+       indicates an inaccessible page. For performance, the reading is first
        done using a buffer (of the size of a page), but once EFAULT is reached,
        it is re-tried byte-by-byte for the last buffer not read successfully.
     */
@@ -545,14 +574,14 @@ static char* findBound(char *from, char *limit, int dir)
    than allocating large arrays on the C stack, both via definition and
    alloca. */
 static SEXP RinitJVM_with_padding(SEXP par, intptr_t padding, char *last) {
-  char dummy[1];
+  volatile char dummy[1];
     /* reduce the risk that dummy will be optimized out */
   dummy[0] = (char) (uintptr_t) &dummy;
   padding -= (last - dummy) * R_CStackDir;
   if (padding <= 0)
     return RinitJVM_real(par, 0);
   else
-    return RinitJVM_with_padding(par, padding, dummy);
+    return RinitJVM_with_padding(par, padding, (char*) dummy);
 }
 
 /* Run RinitJVM with the Java stack workaround */
@@ -702,7 +731,7 @@ static SEXP RinitJVM_jsw(SEXP par) {
       /* the message may be confusing when R_CStackLimit was set to -1
          because the original stack size was too large */
       REprintf("Rjava.init.warning: stack size reduced from unlimited to"
-        " %u bytes after JVM initialization.\n", newlim);
+	       " %lu bytes after JVM initialization.\n", (unsigned long) newlim);
       bigloss = 1;
     } else {
       unsigned lost = (unsigned) (oldlim - newlim);
@@ -739,62 +768,62 @@ REP SEXP RinitJVM(SEXP par) {
 #endif
 }
 
-REP void doneJVM() {
+REP void doneJVM(void) {
   (*jvm)->DestroyJavaVM(jvm);
   jvm = 0;
   eenv = 0;
+  rJava_JVM_state = JVM_STATE_DESTROYED;
 }
 
 /**
  * Initializes the cached values of classes and methods used internally
  * These classes and methods are the ones that are in rJava (RJavaTools, ...)
  * not java standard classes (Object, Class)
- */ 
-REPC SEXP initRJavaTools(){
+ */
+REPC SEXP initRJavaTools(void){
 
 	JNIEnv *env=getJNIEnv();
-	jclass c; 
-	
+	jclass c;
+
 	/* classes */
-	
+
 	/* RJavaTools class */
 	c = findClass(env, "RJavaTools", oClassLoader);
 	if (!c) error("unable to find the RJavaTools class");
 	rj_RJavaTools_Class=(*env)->NewGlobalRef(env, c);
 	if (!rj_RJavaTools_Class) error("unable to create a global reference to the RJavaTools class");
 	(*env)->DeleteLocalRef(env, c);
-	
+
 	/* RJavaImport */
-	c = findClass(env, "RJavaImport", oClassLoader); 
+	c = findClass(env, "RJavaImport", oClassLoader);
 	if (!c) error("unable to find the RJavaImport class");
 	rj_RJavaImport_Class=(*env)->NewGlobalRef(env, c);
 	if (!rj_RJavaImport_Class) error("unable to create a global reference to the RJavaImport class");
 	(*env)->DeleteLocalRef(env, c);
-	
-	
+
+
 	/* methods */
-	
-	mid_RJavaTools_getFieldTypeName  = (*env)->GetStaticMethodID(env, rj_RJavaTools_Class, 
+
+	mid_RJavaTools_getFieldTypeName  = (*env)->GetStaticMethodID(env, rj_RJavaTools_Class,
 		"getFieldTypeName", "(Ljava/lang/Class;Ljava/lang/String;)Ljava/lang/String;");
 	if (!mid_RJavaTools_getFieldTypeName) error("cannot obtain RJavaTools.getFieldTypeName method ID");
-	
-	mid_rj_getSimpleClassNames  = (*env)->GetStaticMethodID(env, rj_RJavaTools_Class, 
+
+	mid_rj_getSimpleClassNames  = (*env)->GetStaticMethodID(env, rj_RJavaTools_Class,
 		"getSimpleClassNames", "(Ljava/lang/Object;Z)[Ljava/lang/String;");
 	if (!mid_rj_getSimpleClassNames) error("cannot obtain RJavaTools.getDimpleClassNames method ID");
-	
-	mid_RJavaImport_getKnownClasses = (*env)->GetMethodID(env, rj_RJavaImport_Class, 
+
+	mid_RJavaImport_getKnownClasses = (*env)->GetMethodID(env, rj_RJavaImport_Class,
 		"getKnownClasses", "()[Ljava/lang/String;");
 	if (!mid_RJavaImport_getKnownClasses) error("cannot obtain RJavaImport.getKnownClasses method ID");
-	
-	mid_RJavaImport_lookup = (*env)->GetMethodID(env, rj_RJavaImport_Class, 
+
+	mid_RJavaImport_lookup = (*env)->GetMethodID(env, rj_RJavaImport_Class,
 		"lookup", "(Ljava/lang/String;)Ljava/lang/Class;");
 	if( !mid_RJavaImport_lookup) error("cannot obtain RJavaImport.lookup method ID");
-	
-	mid_RJavaImport_exists = (*env)->GetMethodID(env, rj_RJavaImport_Class, 
+
+	mid_RJavaImport_exists = (*env)->GetMethodID(env, rj_RJavaImport_Class,
 		"exists", "(Ljava/lang/String;)Z");
 	if( ! mid_RJavaImport_exists ) error("cannot obtain RJavaImport.exists method ID");
 	// maybe add RJavaArrayTools, ...
-	
-	return R_NilValue; 
-}
 
+	return R_NilValue;
+}
